/*
 *	Copyright © 2005–2022 Jan Bobrowski
 *
 *	This program is free software; you can redistribute it and
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mktap.h"

static i32 ungot = -1;

static void uni_unget(i32 c)
{
	ungot = c;
}

static i32 uni_get()
{
	int c;

	if (ungot >= 0) {
		i32 u = ungot;
		ungot = -1;
		return u;
	}

again:
	c = getc(in);
	if (c < 0b11000000) {
		if (c == '\r')
			goto again;
		return c;
	}

	i32 u;
	int n = 1;
	if (c < 0b11100000) {
		u = c & 0b00011111;
	} else {
		n++;
		if (c < 0b11110000)
			u = c & 0b00001111;
		else if (c < 0b11111000) {
			n++;
			u = c & 0b00000111;
		} else
			return c;
	}
	do {
		c = getc(in);
		if (c < 0b10000000 || c >= 0b110000000) {
			if (c >= 0)
				ungetc(c, in);
			return '?';
		}
		u = u<<6 | c & 0b00111111;
	} while (--n);
	return u;
}

extern struct buf data;
static int line_len_pos;

static void append(u8 v)
{
	buf_need(&data, 1);
	data.ptr[data.size++] = v;
}

void start_line(int n)
{
	append(n >> 8); append(n & 0xFF);
	append(0); append(0); // Length will be filled later
	line_len_pos = data.size;
}

static int checkkw();

void end_line()
{
	checkkw();
	append(13);
	int n = data.size - line_len_pos;
	data.ptr[line_len_pos - 2] = n & 0xFF;
	data.ptr[line_len_pos - 1] = n >> 8;
	line_len_pos = -1;
}

static char *stab[] = {
/*2*/	"PI\247FN\250AT\254LN\270IN\277OR\305<=\307>=\310<>\311TO\314IF\372",
/*3*/	"RND\245TAB\255VAL\260LEN\261SIN\262COS\263TAN\264ASN\265ACS\266"
	"ATN\267EXP\271INT\272SQR\273SGN\274ABS\275USR\300NOT\303BIN\304"
	"AND\306CAT\317INK\331OUT\337NEW\346DIM\351REM\352FOR\353LET\361"
	"RUN\367CLS\373",
/*4*/	"ATTR\253VAL$\256CODE\257PEEK\276STR$\301CHR$\302LINE\312THEN\313"
	"STEP\315MOVE\321BEEP\327OVER\336STOP\342READ\343DATA\344LOAD\357"
	"LIST\360NEXT\363POKE\364PLOT\366SAVE\370DRAW\374COPY\377" "GOTO\354"
	"PLAY\244",
/*5*/	"POINT\251ERASE\322MERGE\325PAPER\332FLASH\333LLIST\341GO TO\354"
	"INPUT\356PAUSE\362PRINT\365CLEAR\375" "OPEN#\323GOSUB\355",
/*6*/	"INKEY$\246DEF FN\316FORMAT\320OPEN #\323VERIFY\326CIRCLE\330BRIGHT\334"
	"LPRINT\340BORDER\347GO SUB\355RETURN\376" "CLOSE#\324",
/*7*/	"SCREEN$\252CLOSE #\324INVERSE\335RESTORE\345",
/*8*/	"CONTINUE\350" "SPECTRUM\243",
/*9*/	"RANDOMIZE\371",
};

#define DEF_FN 0316
#define BIN 0304

static int checkkw()
{
	int l = 9;
	for (; l >= 2; l--) {
		if (data.size < l)
			continue;

		unsigned char *s = data.ptr + data.size - l;
		char *p = stab[l - 2];
		if (s[-1] >= 'A' && s[-1] <= 'Z')
			continue;
		do {
			if (memcmp(s, p, l) == 0) {
				int c = p[l] & 0xFF;
				if (s[-1] == ' ')
					s--;
				*s++ = c;
				data.size = s - data.ptr;
				return c;
			}
			p += l + 1;
		} while (*p);
	}
	return 0;
}

static void outE(double v)
{
	append(14);
	if ((u16)v == v) {
		u16 s = v;
		append(0); append(0);
		append(s & 0xFF); append(s >> 8);
		append(0);
	} else {
		int e;
		u32 m = frexp(v, &e) * 65536 * 65536;
		append(e + 128);
		append(m>>24 & 0x7F);
		append(m>>16);
		append(m>>8);
		append(m);
	}
}

static void cvtnum(i32 c, int bin)
{
	double v = 0;

	if (bin) {
		while (c >= '0' && c <= '1' && v < 32768) {
			append(c);
			v = 2*v + (c & 1);
			c = uni_get();
		}
		goto out;
	}

	if (c == '.')
		goto real;
	for (;;) {
		append(c);
		v = 10*v + c - '0';
		c = uni_get();
		if (c < '0' || c > '9')
			break;
	}
	if (c == '.') {
		double m;
real:
		m = 1;
		for (;;) {
			append(c);
			c = uni_get();
			if (c < '0' || c > '9')
				break;
			m /= 10;
			v += m * (c - '0');
		}
	}

out:
	uni_unget(c);
	outE(v);
}

static u8 char_map(i32 c)
{
	if (c < 0x7f)
		return c;
	switch (c) {
	case 0x00A9: c = 127; break; // ©
	case 0x00A3: c = 96; break; // £
	case 0x2191: c = '^'; break; // ↑
	case 0x00A0: c = 0x80; break;
	case 0x259D: c = 0x81; break; // ▝
	case 0x2598: c = 0x82; break; // ▘
	case 0x2580: c = 0x83; break; // ▀
	case 0x2597: c = 0x84; break; // ▗
	case 0x2590: c = 0x85; break; // ▐
	case 0x259A: c = 0x86; break; // ▚
	case 0x259C: c = 0x87; break; // ▜
	case 0x2596: c = 0x88; break; // ▖
	case 0x259E: c = 0x89; break; // ▞
	case 0x258C: c = 0x8a; break; // ▌
	case 0x259B: c = 0x8b; break; // ▜
	case 0x2584: c = 0x8c; break; // ▄
	case 0x259F: c = 0x8d; break; // ▟
	case 0x2599: c = 0x8e; break; // ▙
	case 0x2588: c = 0x8f; break; // █
	default:
		if (c >= 0xe000 && c <= 0xe000 + 'U' - 'A') {
			c += 0x90 - 0xe000;
			break;
		}
		c = '?';
	}
	return c;
}

static int linenr = -1;
static void def_fn(); // special case

static int cvtline(char *prefix)
{
	i32 c;
	int g;

	c = uni_get();
	if (c < 0)
		return 0;
	if (c == 0xfeff)
		c = uni_get();

	if (prefix) { // TODO We're ignoring unicode
		while (*prefix) {
			if (c != *prefix++)
				goto skip_rest;
			c = uni_get();
			if (c < 0)
				return 0;
		}
	}

	while (c <= ' ') {
		if (c == '\n')
			return 1;
		c = uni_get();
		if (c < 0)
			return 0;
	}

	if (c == '#') { // Comment
		do {
			c = uni_get();
			if (c < 0)
				return 0;
skip_rest:;
		} while (c != '\n');
		return 1;
	}

	linenr++;
	if (c >= '0' && c <= '9') {
		linenr = 0;
		do {
			linenr = 10*linenr + (c - '0');
			c = uni_get();
		} while (c >= '0' && c <= '9');
	}
	start_line(linenr);

	if (c != '>')
		uni_unget(c); // Skip line marker

	for (;;) {
		c = uni_get();
		if (c < 0)
			goto fail;
		if (c == '"') {
			checkkw();
			g = 0;
			do {
				append(c);
get_next:
				c = uni_get();
				if (c == 10 || c < 0) {
					uni_unget(c);
					goto fail;
				}
				if (c == '`') {
					g ^= 0xe000 - 'A';
					goto get_next;
				}
				if (g) {
					if (c >= 'A' && c <= 'U')
						c += g;
					else if (c == 'c')
						c = 0x00A9; // ©
					else if (c == '#')
						c = 0x00A3; // £
					g = 0;
				}
				c = char_map(c);
			} while (c != '"');
		}
		if (c >= '0' && c <= '9' || c == '.') {
			checkkw();
			cvtnum(c, 0);
			continue;
		}
		if (c < 'A' && c != '$' || c > 'Z') {
			int kw = checkkw();
			switch (kw) {
			case DEF_FN:
				if (c != ' ')
					uni_unget(c);
				def_fn();
				continue;
			case BIN:
				if (c == ' ')
					c = uni_get();
				cvtnum(c, BIN);
				continue;
			default:
				if (c == ' ')
					continue;
			case 0:
				break;
			}
		}
		if (c == 10)
			break;
		append(c);
	}
	end_line();
	return 1;
fail:
	end_line();
	return 0;
}

static void def_fn()
{
	const char *term = "\n:)";
	i32 c;
	do {
		c = uni_get();
		if (c < 0 || strchr(term, c))
			goto ret;
		append(c);
	} while (c != '(');
	for (;;) {
		c = uni_get();
		if (c < 0 || strchr(term, c))
			goto ret;
		append(c);
		if ((c < 'a' || c > 'z') && (c < 'A' || c > 'Z'))
			continue;
		c = uni_get();
		if (c == '$')
			append(c);
		else {
			if (c < 0)
				return;
			uni_unget(c);
		}
		outE(0);
	}
ret:
	uni_unget(c);
}

void read_basic(char *prefix)
{
	while(cvtline(prefix))
		;
}
