/*
 *	Copyright © 2005–2022 Jan Bobrowski
 *
 *	This program is free software; you can redistribute it and
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if !defined _WIN32 && !defined _DOS
#include <err.h>
#include <unistd.h>
#else
#include <fcntl.h>
#include <stdarg.h>
#ifdef __ia16__
#include <unistd.h>
#endif

void warnx(char *f, ...)
{
	va_list a; va_start(a, f);
	vfprintf(stderr, f, a); va_end(a);
	fprintf(stderr, "\n");
}

void errx(int s, char *f, ...)
{
	va_list a; va_start(a, f);
	vfprintf(stderr, f, a);
	va_end(a);
	fprintf(stderr, "\n");
	exit(s);
}

#endif

#include "mktap.h"

#pragma clang diagnostic ignored "-Wshift-op-parentheses"

void buf_need(struct buf *buf, unsigned more)
{
	unsigned need = buf->size + more;
	if (need <= buf->allocated)
		return;
	need = need+010000 & ~07777;
	buf->ptr = realloc(buf->ptr, need);
	if (!buf->ptr)
		err(1, 0);
	buf->allocated = need;
}

void buf_append(struct buf *buf, void *src, unsigned size)
{
	buf_need(buf, size);
	memcpy(buf->ptr + buf->size, src, size);
	buf->size += size;
}

struct buf data = {0, 0, 0};

enum block_marker {
	MarkHeader = 0x00,
	MarkData = 0xff
};

enum tap_type {
	Program = 0,
	NumberArray = 1,
	CharacterArray = 2,
	Bytes = 3,
	TypeUnknown = -1
};

typedef struct u16le { u8 l, h; } u16le;

struct header {
	u8 type;
	u8 name[10];
	u16le length, start, prog;
};

#define put16(D,V) ((D).l=(V), (D).h=(V)>>8)

static void read_into_data(FILE *f)
{
	for (;;) {
		int v;
		buf_need(&data, 1024);
		v = fread(data.ptr + data.size, 1, 1024, f);
		if (v == 0)
			break;
		data.size += v;
	}
}

enum put_in_rem {
	NoREMAddr,
	CallREM,
	REMAddrInVar
};

#define XX 0x00 // Mark bytes to be filled later

static u8 LINE_0[] = {
	0x00, 0x00, XX, XX, // Line 0
};

static u8 RANDOMIZE_USR[] = {
	0xf9, 0xc0
};

static u8 GET_ADDR_REM[] = {
	'(', '0', 0x0e, 0x00, 0x00, XX, XX, 0x00, // (X
	'+', 0xbe, '0', 0x0e, 0x00, 0x00, 0x53, 0x5c, 0x00, // + PEEK 23635
	'+', '0', 0x0e, 0x00, 0x00, 0x00, 0x01, 0x00, // +256
	'*', 0xbe, '0', 0x0e, 0x00, 0x00, 0x54, 0x5c, 0x00, // * PEEK 23636
	')', ':', 0xea
};
#define VAL_TO_ADD_OFS (5 - sizeof GET_ADDR_REM)

static void put_in_rem(FILE *f, enum put_in_rem variant, char *variable)
{
	static u8 REM = 0xea;
	static u8 LET = 0xf1;
	static u8 EQ = '=';
	static u8 CR = 13;
	unsigned line_pos = data.size;
	buf_append(&data, LINE_0, sizeof LINE_0);
	if (variant != NoREMAddr) {
		if (variant == REMAddrInVar) {
			buf_append(&data, &LET, sizeof LET);
			buf_append(&data, variable, strlen(variable));
			buf_append(&data, &EQ, sizeof EQ);
		} else if (variant == CallREM) {
			buf_append(&data, RANDOMIZE_USR, sizeof RANDOMIZE_USR);
		}
		buf_append(&data, GET_ADDR_REM, sizeof GET_ADDR_REM);
		unsigned o = data.size + VAL_TO_ADD_OFS;
		data.ptr[o + 0] = data.size;
		data.ptr[o + 1] = data.size >> 8;
	} else {
		buf_append(&data, &REM, sizeof REM);
	}
	read_into_data(f);
	buf_append(&data, &CR, sizeof CR);
	unsigned n = data.size - line_pos - 4;
	data.ptr[line_pos + 2] = n;
	data.ptr[line_pos + 3] = n >> 8;
}

static void save_block(enum block_marker marker, u8 *d, int l)
{
	u8 x = marker;
	int n;

	putchar(l+2);
	putchar(l+2 >> 8);

	putchar(x);
	n = fwrite(d, 1, l, stdout);
	if (n < l)
		errx(1, "fwrite: %d < %d", n, l);

	for (n = 0; n < l; n++)
		x ^= d[n];
	putchar(x);
}

static void usage(FILE *o, int requested)
{
	if (requested)
		fprintf(o,
			"MkTAP by Jan Bobrowski\n"
			"version: "STR(VERSION)", compiled: "__DATE__"\n"
		);
	fprintf(o,
		"mktap [opts] name [start] <input >output.tap\n"
	);
	if (requested)
		fprintf(o,
			" -b       Convert text file to Basic program\n"
			" -p str   Consider only lines starting with str (if -b)\n"
			" -t num   File type (0:basic, 3:bytes)\n"
			" -r       Headless (data block only)\n"
			" -rr      Raw data (not a tape)\n"
			" -R file  Data to put into REM (if -b)\n"
			" -U       Put into REM statement and call it\n"
			" -V var   Put into REM statement and set var to its address\n"
			"In Basic `A..`U introduce UDG's\n"
		);
	else
		fprintf(o, "mktap -h\n");
	exit(!requested);
}

int main(int argc, char *argv[])
{
	int raw = 0;
	int basic = 0;
	char *prefix = 0;
	enum tap_type type = TypeUnknown;
	enum block_marker marker = MarkData;
	u16 start;
	enum put_in_rem rem_variant = NoREMAddr;
	char *rem_data = 0;
	char *variable = 0;
	struct header h;

	for (;;) switch (getopt(argc, argv, "?hrt:bp:R:V:U")) {
	case EOF: goto endopt;
	case 'r': raw++; break;
	case 'b': basic = 1; break;
	case 'p': prefix = optarg; break;
	case 't': marker = type = atoi(optarg); break;
	case 'R': rem_data = optarg; break;
	case 'V': variable = optarg; rem_variant = REMAddrInVar; break;
	case 'U': rem_variant = CallREM; break;
	case '?':
		if (optopt)
			usage(stderr, 0);
	case 'h':
		usage(stdout, 1);
	default:
		usage(stderr, 0);
	}
endopt:

	if (optind == argc)
		usage(isatty(1) ? stdout : stderr, 0);

	{
		char *name = argv[optind++];
		int l = strlen(name);
		if (l > sizeof h.name)
			l = sizeof h.name;
		memset(h.name, ' ', sizeof h.name);
		memcpy(h.name, name, l);
	}

	start = 0xffff;
	if (optind < argc) {
		start = atoi(argv[optind++]);
		if (optind < argc)
			warnx("Too many args: %s", argv[optind]);
	}

#ifdef _WIN32
	_setmode(_fileno(stdout), _O_BINARY);
#endif

	if (rem_data && !basic)
		errx(1, "R option requires explicit Basic (-b)");

	if (!basic && !rem_data && rem_variant == NoREMAddr) {
		if (start < 0)
			start = 0;
		read_into_data(in);
		if (type == TypeUnknown)
			type = Bytes;
	} else {
		if (rem_data) {
			FILE *f = fopen(rem_data, "rb");
			if (!f)
				errx(1, "Can't open %s", rem_data);
			put_in_rem(f, rem_variant, variable);
			read_basic(prefix);
			fclose(f);
		} else if (!basic) {
			put_in_rem(in, rem_variant, variable);
		} else {
			read_basic(prefix);
		}
		if (type == TypeUnknown)
			type = Program;
		basic = 1;
	}

	switch (raw) {
	case 0:
		h.type = type;
		put16(h.length, data.size);
		put16(h.start, start);
		put16(h.prog, basic ? data.size : 0);
		save_block(MarkHeader, (u8 *)&h, sizeof h);
		save_block(MarkData, data.ptr, data.size);
		break;
	case 1:
		save_block(marker, data.ptr, data.size);
		break;
	default:
		fwrite(data.ptr, 1, data.size, stdout);
		break;
	}

	return 0;
}
